__all__: list[str] = []

# Functions
def addSamplesDataSearchPath(path: str) -> None: ...

def addSamplesDataSearchSubDirectory(subdir: str) -> None: ...

def findFile(relative_path: str, required: bool = ..., silentMode: bool = ...) -> str: ...

def findFileOrKeep(relative_path: str, silentMode: bool = ...) -> str: ...


