__all__: list[str] = []

# Functions
def getIppVersion() -> str: ...

def setUseIPP(flag: bool) -> None: ...

def setUseIPP_NotExact(flag: bool) -> None: ...

def useIPP() -> bool: ...

def useIPP_NotExact() -> bool: ...


